from emod_api import schema_to_class as s2c
from emod_api import campaign

from typing import List, Union
from enum import Enum
import warnings
import re


# This value represents the default maximum age in years. It is set to a large value (1000 years) to cover all possible
# age ranges in the simulation.
MAX_AGE_YEARS = 1000


class TargetGender(Enum):
    ALL = "All"
    MALE = "Male"
    FEMALE = "Female"


class TargetDemographicsConfig:
    """
    A class that is used to configure the Demographics_Coverage, Target_Demographic, Target_Age_Min, Target_Age_Max,
    Target_Gender and Target_Residents_Only in the coordinator class. Please refer to emodpy.campaign.distributor
    for its usage.

    Args:
        demographics_coverage (float, optional):
            - The fraction of the demographic covered by the event.
            - Defaults to 1.0.

        target_age_min (float, optional):
            - The minimum age targeted by the event, in years.
            - Defaults to 0.

        target_age_max (float, optional):
            - The maximum age targeted by the event, in years.
            - Defaults to MAX_AGE_YEARS.

        target_gender (Gender, optional):
            - A member of the Gender enum indicating the gender targeted by the event.
            - The function accepts TargetGender.ALL, TargetGender.MALE, or TargetGender.FEMALE as valid inputs.
            - Defaults to Gender.ALL.

        target_residents_only (bool, optional):
            - When set to true, the intervention is only distributed to individuals that began the simulation in the node (i.e. those that claim the node as their residence).
            - You can use the MigrateIndividuals function to modify an individual's HOME node. For more details on MigrateIndividuals, refer to the Emod documentation: :doc:`emod-hiv:parameter-campaign-individual-migrateindividuals`
            - Defaults to False.

    Examples:
        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import TargetDemographicsConfig, TargetGender
        >>> from emodpy.campaign.distributor import add_intervention_scheduled
        >>> demographics_config = TargetDemographicsConfig(demographics_coverage=0.5, target_age_min=10,
        >>>                                                target_age_max=20, target_gender=TargetGender.FEMALE)
        >>> add_intervention_scheduled(demographics_config=demographics_config, ...)
    """
    class _TargetDemographic(Enum):
        EVERYONE = "Everyone"
        EXPLICIT_AGE_RANGES = "ExplicitAgeRanges"
        EXPLICIT_AGE_RANGES_AND_GENDER = "ExplicitAgeRangesAndGender"
        EXPLICIT_GENDER = "ExplicitGender"
        _POSSIBLE_MOTHERS = "PossibleMothers"  # TODO: Not yet implemented
        _EXPLICIT_DISEASE_STATE = "ExplicitDiseaseState"  # TODO: Not yet implemented

    def __init__(self, demographics_coverage: Union[float, None] = 1.0, target_age_min: float = 0,
                 target_age_max: float = MAX_AGE_YEARS, target_gender: TargetGender = TargetGender.ALL,
                 target_residents_only: bool = False):
        self.demographics_coverage = demographics_coverage
        self.target_age_min = target_age_min
        self.target_age_max = target_age_max
        self.target_gender = target_gender
        self.target_residents_only = target_residents_only

    def _set_target_demographics(self, campaign_object: s2c.ReadOnlyDict):
        """
        A function that configure the Target_Age_Min, Target_Age_Max, Target_Gender and Target_Residents_Only for the
        campaign_object and set Target_Demographic based on the values of the first 3 parameters.

        This function is a private function that is used by the coordinator class and should not be called directly by
        user code.

        Args:
            campaign_object(s2c.ReadOnlyDict): an EventCoordinator object.

        Returns:
            This function does not return any value; instead, it modifies the campaign_object.

        """
        if self.demographics_coverage is not None:
            campaign_object.Demographic_Coverage = self.demographics_coverage
        if self.target_residents_only is not None:
            campaign_object.Target_Residents_Only = self.target_residents_only

        # Target_Demographic is 'ExplicitAgeRangesAndGender'
        if self.target_gender != TargetGender.ALL and (self.target_age_min > 0 or self.target_age_max < MAX_AGE_YEARS):
            campaign_object.Target_Age_Min = self.target_age_min
            campaign_object.Target_Age_Max = self.target_age_max
            campaign_object.Target_Gender = self.target_gender.value
            campaign_object.Target_Demographic = self._TargetDemographic.EXPLICIT_AGE_RANGES_AND_GENDER.value
        # Target_Demographic is 'ExplicitGender'
        elif self.target_gender != TargetGender.ALL:
            campaign_object.Target_Gender = self.target_gender.value
            campaign_object.Target_Demographic = self._TargetDemographic.EXPLICIT_GENDER.value
        # Target_Demographic is 'ExplicitAgeRanges'
        elif self.target_age_min > 0 or self.target_age_max < MAX_AGE_YEARS:
            campaign_object.Target_Age_Min = self.target_age_min
            campaign_object.Target_Age_Max = self.target_age_max
            campaign_object.Target_Demographic = self._TargetDemographic.EXPLICIT_AGE_RANGES.value
        # Set Target_Demographic to 'Everyone' by default
        else:
            campaign_object.Target_Demographic = self._TargetDemographic.EVERYONE.value


class RepetitionConfig:
    """
    A class that is used to configure the number of times the intervention event will occur by setting the
    Number_Repetitions, Timesteps_Between_Repetitions in the Event_Coodinator.

    It's used with emodpy.campaign.distributor.add_intervention_scheduled function.

    Args:
        number_repetitions (int, optional):
            - The number of times the event will occur. If the value is 1, it implies that there will be no repetitions and the event will occur only once.
            - This parameter accepts integer values that are greater than or equal to 1.
            - This argument is ignored if infinite_repetitions is set to True.
            - Defaults to 1.

        timesteps_between_repetitions (int, optional):
            - The number of timesteps between repetitions.
            - Used in conjunction with number_repetitions or infinite_repetitions.
            - If number_repetitions is greater than 1 or infinite_repetitions is set to True, indicating that the event will repeat, timesteps_between_repetitions should be a positive integer interval.
            - Defaults to None.

        infinite_repetitions (bool, optional):
            - If set to True, the event will repeat indefinitely.
            - If True, the number_repetitions argument is ignored and timesteps_between_repetitions must be set to a positive integer.
            - Defaults to False.

    Raises:
        ValueError: if timesteps_between_repetitions is undefined when number_repetitions is used.

    Examples:
        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import RepetitionConfig
        >>> from emodpy.campaign.distributor import add_intervention_scheduled
        >>> repetition_config = RepetitionConfig(number_repetitions=2, timesteps_between_repetitions=365)
        >>> add_intervention_scheduled(repetition_config=repetition_config, ...)

    """
    def __init__(self, number_repetitions: int = 1, timesteps_between_repetitions: int = None,
                 infinite_repetitions: bool = False):
        self.number_repetitions = number_repetitions
        self.timesteps_between_repetitions = timesteps_between_repetitions
        self.infinite_repetitions = infinite_repetitions
        self._verify_repetitions()
        # Set number_repetitions to -1 if infinite_repetitions is set to True
        self.number_repetitions = -1 if self.infinite_repetitions else self.number_repetitions

    def _verify_repetitions(self):
        """
        A private function that verifies the values of number_repetitions, timesteps_between_repetitions and
        infinite_repetitions.
        """
        if self.infinite_repetitions:
            if self.number_repetitions != 1:  # if number_repetitions is set to a value other than 1(default value)
                warnings.warn("The number_repetitions is ignored when infinite_repetitions is set to True.")
            self._verify_timesteps()
        else:
            if self.number_repetitions < 0:  # if number_repetitions is set to a negative value
                raise ValueError(f"number_repetitions is set to a negative value: {self.number_repetitions}, "
                                 f"please set it to a positive integer.")
            elif self.number_repetitions in [0, 1]:
                warnings.warn("number_repetitions is set to 0 or 1, the event will not be repeated.")
            elif self.number_repetitions > 1:
                self._verify_timesteps()

    def _verify_timesteps(self):
        if self.timesteps_between_repetitions is None:
            raise ValueError("timesteps_between_repetitions must be set when number_repetitions is greater than 1 or "
                             "infinite_repetitions is set to True..")
        # timesteps_between_repetitions should be a positive integer
        if self.timesteps_between_repetitions <= 0:
            raise ValueError(f"timesteps_between_repetitions is set to a non positive value: "
                             f"{self.timesteps_between_repetitions}, please set it to a positive integer.")

    def _set_repetitions(self, campaign_object: s2c.ReadOnlyDict):
        """
        A function that configure the Number_Repetitions and Timesteps_Between_Repetitions for the campaign_object.
        This function is a private function that is used by the coordinator class and should not be called directly by
        user code.

        Args:
            campaign_object: an EventCoordinator object.

        Returns:
            This function does not return any value; instead, it modifies the campaign_object.

        """
        if self.number_repetitions:
            campaign_object.Number_Repetitions = self.number_repetitions
        if self.timesteps_between_repetitions:
            campaign_object.Timesteps_Between_Repetitions = self.timesteps_between_repetitions


class PropertyRestrictions:
    """
    A class that is used to configure the individual property restrictions and node property restrictions in the
    campaign object.

    Please refer to the Emod documentation for NodeProperties and IndividualProperties parameters for more
    information:

    - HIV Emod: :doc:`emod-hiv:parameter-demographics`
    - Malaria Emod: :doc:`emod-malaria:parameter-demographics`

    Args:
        individual_property_restrictions (List[List[str]], optional):
            - A 2D list contains lists of the IndividualProperty key:value pairs.
            - These are defined in the demographics.
            - Individuals must have these properties to be targeted by the intervention.
            - This parameter allows you to specify AND and OR combinations of key:value pairs. Please see example 1 and 2 for more information.
            - You can either use this parameter or node_property_restrictions, but not both.
            - Defaults to None.

        node_property_restrictions (List[List[str]], optional):
            - A 2D list contains lists of the NodeProperty key:value pairs.
            - These are defined in the demographics.
            - Nodes must have these properties to be targeted by the intervention.
            - You can specify AND and OR combinations of key:value pairs with this parameter. Please see example 3 section for more information.
            - You can either use this parameter or individual_property_restrictions, but not both.
            - Defaults to None.

    Raises:
        Warnings: if both individual_property_restrictions and node_property_restrictions are not defined.
        ValueError: if both individual_property_restrictions and node_property_restrictions are defined.
        ValueError: if individual_property_restrictions or node_property_restrictions is not a 2D list(List[List[str]]).
        ValueError: if the elements in the inner list are not strings that represent dictionaries key:value pairs with at least one alphanumeric character before and after ':'.

    Examples:
        Example 1: This example demonstrates how to specify individual restrictions for targeting specific groups of people who are high risk AND whose InterventionStatus is ARTStaging.

        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import PropertyRestrictions
        >>> from emodpy.campaign.distributor import add_intervention_scheduled
        >>> property_restrictions = PropertyRestrictions(individual_property_restrictions=[["Risk:HIGH", "InterventionStatus:ARTStaging"]])
        >>> add_intervention_scheduled(property_restrictions=property_restrictions, ...)
        >>> # the result json should look like this:
        >>> # {
        >>> #          "Property_Restrictions_Within_Node": [
        >>> #            {
        >>> #              "Risk": "HIGH",
        >>> #              "InterventionStatus": "ARTStaging"
        >>> #            }
        >>> #          ]
        >>> # }


        Example 2: This example demonstrates how to specify individual restrictions for targeting specific groups of people. In this case, we are targeting individuals whose InterventionStatus is set to ARTStaging and who have either HIGH or MEDIUM risk behavior. In other words, we aim to target individuals who meet either of the following conditions:

                    1. "InterventionStatus is set to ARTStaging and Risk is set to HIGH"
                    2. "InterventionStatus is set to ARTStaging AND Risk is set to MEDIUM"

        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import PropertyRestrictions
        >>> from emodpy.campaign.distributor import add_intervention_scheduled
        >>> property_restrictions_within_node = PropertyRestrictions(
        >>>                                       individual_property_restrictions=[
        >>>                                                 ["Risk:HIGH", "InterventionStatus:ARTStaging"],
        >>>                                                 ["Risk:MEDIUM", "InterventionStatus:ARTStaging"]])
        >>> add_intervention_scheduled(property_restrictions=property_restrictions, ...)
        >>> # the result json should look like this:
        >>> # {
        >>> #          "Property_Restrictions_Within_Node": [
        >>> #            {
        >>> #              "Risk": "HIGH",
        >>> #              "InterventionStatus": "ARTStaging"
        >>> #            },
        >>> #            {
        >>> #              "Risk": "MEDIUM",
        >>> #              "InterventionStatus": "ARTStaging"
        >>> #            }
        >>> #          ]
        >>> # }

        Example 3: This example demonstrates how to use 'node_property_restrictions' to specify the NodeProperty. In this case, we are targeting nodes that meet either of the following conditions:

                    1. "Risk is set to MEDIUM and Place is set to URBAN"
                    2. "Risk is set to LOW and Place is set to RURAL"

        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import PropertyRestrictions
        >>> from emodpy.campaign.distributor import add_intervention_scheduled
        >>> property_restrictions = PropertyRestrictions(
        >>>                            node_property_restrictions=[
        >>>                                                 ["Risk:MEDIUM", "Place:URBAN"],
        >>>                                                 ["Risk:LOW", "Place:RURAL"]])
        >>> add_intervention_scheduled(property_restrictions=property_restrictions, ...)
        >>> # the result json should look like this:
        >>> # {
        >>> #          "Node_Property_Restrictions": [
        >>> #            {
        >>> #              "Risk": "MEDIUM",
        >>> #              "Place": "URBAN"
        >>> #            },
        >>> #            {
        >>> #              "Risk": "LOW",
        >>> #              "Place": "RURAL"
        >>> #            }
        >>> #          ]
        >>> # }

    """
    def __init__(self, individual_property_restrictions: List[List[str]] = None,
                 node_property_restrictions: List[List[str]] = None):
        self.individual_property_restrictions = individual_property_restrictions
        self.node_property_restrictions = node_property_restrictions
        self._verify_property_restrictions()

    def _verify_property_restrictions(self):
        if not self.individual_property_restrictions and not self.node_property_restrictions:
            warnings.warn("No property restrictions are provided.")
        elif self.individual_property_restrictions and self.node_property_restrictions:
            raise ValueError("Both individual_property_restrictions and node_property_restrictions are provided. "
                             "Please provide only one of them.")
        elif self.individual_property_restrictions:
            self._validate_restrictions(self.individual_property_restrictions, "individual_property_restrictions")
        else:  # if self.node_property_restrictions:
            self._validate_restrictions(self.node_property_restrictions, "node_property_restrictions")

    @staticmethod
    def _validate_restrictions(restrictions_list, restrictions_name):
        # check if restrictions_list is a 2D list(List[List[str]])
        if not isinstance(restrictions_list, list) or not all(isinstance(restriction, list) for restriction in
                                                              restrictions_list):
            raise ValueError(f"The {restrictions_name} should be a 2D list(List[List[str]]). "
                             f"Got {restrictions_list}. Please check your input.")
        for restrictions in restrictions_list:
            for restriction in restrictions:
                if not re.match(r"\s*\w+\s*:\s*\w+\s*", restriction):
                    raise ValueError(f"The elements in the inner list of {restrictions_name} should be strings that "
                                     f"represent dictionaries key:value pairs with at least one alphanumeric character "
                                     f"before and after ':'. Got '{restriction}'. Please check your input.")

    def _set_property_restrictions(self, campaign_object: s2c.ReadOnlyDict):
        """
        A function that configure the Property_Restrictions_Within_Node and Node_Property_Restrictions
        for the campaign_object. This function is a private function that is used by the coordinator class and should
        not be called directly by user code.
        Args:
            campaign_object: An EventCoordinator or Intervention object.

        Returns:
            This function does not return any value; instead, it modifies the campaign_object.

        """
        def parse_to_dict(list_of_lists):
            """
            A helper function that parse the list of lists to a list of dictionaries of key-value-pairs where each dict
            can only contain a given key once.
            Args:
                list_of_lists:

            Returns:
                list of dictionaries of key-value-pairs where each dict can only contain a given key once.

            """
            result = []
            for inner_list in list_of_lists:
                dict_item = {}
                for item in inner_list:
                    key, value = item.split(':')
                    dict_item[key.strip()] = value.strip()
                result.append(dict_item)
            return result

        if self.individual_property_restrictions:
            campaign_object.Property_Restrictions_Within_Node = parse_to_dict(self.individual_property_restrictions)
            if hasattr(campaign_object, 'Property_Restrictions'):
                campaign_object.Property_Restrictions = []
        if self.node_property_restrictions:
            campaign_object.Node_Property_Restrictions = parse_to_dict(self.node_property_restrictions)


class ValueMap:
    """
    Create an object that is used to configure the Times and Values for InterpolatedValueMap in the campaign_object.

    The ValueMap object is initialized with two lists: times and values. The times list represents specific points in
    time (in years) and the values list represents the corresponding values at those times. The values that are
    interpolated linearly between the times specified in the times list.

    The number of elements in the times and values lists should be the same, and the times list should be in ascending
    order.

    Args:
        camp (campaign):
            - The campaign object to which the value map will be added.

        times (List[float]):
            - A list of times (in years) at which the value changes.
            - The times should be in ascending order.

        values (List[float]):
            - A list of values that correspond to the times.
            - The list of values should have the same length as times.

    Examples:
        In this example, we create a ValueMap object with the following times and values:
        [2005, 2010, 2012, 2013] and [10, 20, 33, 40] respectively. The ValueMap object is then used to create an
        InterpolatedValueMap object in the campaign object. The value at time 2005 is 10, at time 2010 is 20, at time 2012
        is 33, and at time 2013 is 40. The values between these times are interpolated linearly. For example, the value
        at time 2011 is 26.5, which is the average of the values at times 2010 and 2012. The value at time 2006 is 12,
        which is calculated by linear interpolation between the values at times 2005 and 2010,
        (20-10)*(2006-2005)/(2010-2005) + 10 , and so on.

        >>> # replace emodpy with emodpy_hiv or emodpy_malaria based on the disease you are working on.
        >>> from emodpy.campaign.common import ValueMap
        >>> camp = emod_api.campaign
        >>> camp.schema_path = "path/to/schema"
        >>> times = [2005, 2010, 2012, 2013]
        >>> values = [10, 20, 33, 40]
        >>> value_map = ValueMap(camp=camp, times=times, values=values)
    """
    def __init__(self, camp: campaign, times: List[float], values: List[float]):
        if len(times) != len(values):
            raise ValueError("The length of times and values should be the same.")
        if times != sorted(times):
            raise ValueError("The times should be in ascending order.")
        self._times = times
        self._values = values
        self.map = self._new_value_map(camp)

    def _new_value_map(self, camp) -> s2c.ReadOnlyDict:
        value_map = s2c.get_class_with_defaults("idmType:InterpolatedValueMap", camp.schema_path)
        value_map.Times = self._times
        value_map.Values = self._values
        return value_map
